;; jde-introspect.el -- Introspect class source files
;;
;; $Id$
;;
;; Author: Sandip Chitale

;; Keywords: java, introspect classes

;; Copyright (C) 2003 Sandip Chitale

;; This package follows the GNU General Public Licence (GPL), see the
;; COPYING file that comes along with GNU Emacs. This is free software,
;; you can redistribute it and/or modify it under the GNU GPL terms.
;;
;; Java is a registered trademark of Sun Microsystem, Inc.
;;
;;; Commentary:

;; This is packages that works with the
;; Java Development Environment (JDE) for Emacs. See the
;; JDE User's Guide for more information about JDE.

;; This package allows introspection of the type at point.
;; It shows the info about the CTORs, Methods and fields of
;; type in a buffer. The types in CTORs, methods and fields
;; are hyperlinks.

;;; Usage:

;; Just put the following in your .emacs
;;
;; (require 'jde-introspect)
;;
;; Now put the point anywhere in Java buffer where you would
;; normally invoke jde-complete functions. Then type
;; (control c) (control v) (/) to see the java typeinfo
;; buffer.

;; Known bugs/problems :
;;
;; JDEE Info:
;;
;; The latest version of the JDE is available at
;; <URL:http://jde.sunsite.dk>.
;; <URL:http://www.geocities.com/SiliconValley/Lakes/1506/>

;; Please send any comments, bugs to
;; Sandip Chitale (sandipchitale@yahoo.com)

(require 'jde)

(defun jde-introspect-javatype-at-point ()
  "Shows the introspection based info about the type where current point is.
 This function needs the same requirements to work as the
 field/method-completion-feature in JDE \(see
`jde-complete-at-point')!. The source-file is searched first in
`jde-sourcepath', then in `jde-global-classpath', then in
$CLASSPATH, then in current-directory."
  (interactive)
  (if (jde-open-functions-exist)
      (let* ((thing-of-interest (thing-at-point 'symbol))
             (pair (save-excursion (end-of-thing 'symbol)
                                   (jde-parse-java-variable-at-point)))
             (class-to-open (jde-open-get-class-to-open
                             pair thing-of-interest))
	     )
        (if (and class-to-open (stringp class-to-open))
            (jde-introspect-javatype class-to-open)
          (error "Can not parse the thing at point!")))
    (message "You need JDE >= 2.2.6 and Senator for using this feature!")))

(define-key jde-mode-map [(control c) (control v) (/)]         'jde-introspect-javatype-at-point)  

(defun jde-introspect-javatype (javatype)
  "Introspect the specified."
  (interactive "sJava Type name : ")
  (if javatype     
      (bsh-eval-r (concat "javatypeinfo(\"" javatype "\");"))
    )
  )

(defun jde-introspect-javatype-by-xref-at-point ()
  "Introspect the javatype based on the xref at point."
  (interactive)
  (let ((javatype (get-text-property (point) 'javatype)))
    (if javatype
	(jde-introspect-javatype javatype)
      )
    )
  )

(defun jde-open-source-javatype-by-xref-at-point ()
  "Introspect the javatype based on the xref at point."
  (interactive)
  (let ((javatype (get-text-property (point) 'javatype-for-source)))
    (if javatype
	(jde-find-class-source javatype)
      )
    )
  )

(defun jde-help-javatype-by-xref-at-point ()
  "Introspect the javatype based on the xref at point."
  (interactive)
  (let ((javatype (get-text-property (point) 'javatype-for-source)))
    (if javatype
	(jde-help-symbol-internal javatype)
      )
    )
  )

;; For tabbing through buffer.
(defun jde-introspect-next-ref ()
  "Find the next help cross-reference in the buffer."
  (interactive)
  (let (pos)
    (while (not pos) 
      (if (get-text-property (point) 'javatype) ; move off reference
	   (goto-char (or (next-single-property-change (point) 'javatype)
                          (point))))
      (cond ((setq pos (next-single-property-change (point) 'javatype))
	     (if pos (goto-char pos)))
	    ((bobp)
	     (message "No cross references in the buffer.")
	     (setq pos t))
	    (t				; be circular
	     (goto-char (point-min)))))))

(defun jde-introspect-previous-ref ()
  "Find the previous help cross-reference in the buffer."
  (interactive)
  (let (pos)
    (while (not pos) 
      (if (get-text-property (point) 'javatype) ; move off reference
	  (goto-char (or (previous-single-property-change (point) 'javatype)
                         (point))))
      (cond ((setq pos (previous-single-property-change (point) 'javatype))
	     (if pos (goto-char pos)))
	    ((bobp)
	     (message "No cross references in the buffer.")
	     (setq pos t))
	    (t				; be circular
	     (goto-char (point-max)))))))
(provide 'jde-introspect)

;; $Log$
