;;; jde-javadoc-pretty.el --- perform pretty operations on javadoc.

;; $Id: jde-javadoc-pretty.el,v 1.5 2003/01/10 05:57:59 burton Exp $

;; Copyright (C) 1997-2000 Free Software Foundation, Inc.

;; Author: Kevin A. Burton (burton@openprivacy.org)
;; Maintainer: Kevin A. Burton (burton@openprivacy.org)
;; Location: http://relativity.yi.org
;; Keywords:
;; Version: 1.0.0

;; This file is [not yet] part of GNU Emacs.

;; This program is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free Software
;; Foundation; either version 2 of the License, or any later version.
;;
;; This program is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License along with
;; this program; if not, write to the Free Software Foundation, Inc., 59 Temple
;; Place - Suite 330, Boston, MA 02111-1307, USA.

;;; Commentary:
;;
;; This package provdies support for decorating and cleansing javadoc within a
;; Java buffer.  The point is that sometime you would lik to write a long
;; javadoc entry and LOT have to worry about its formating.
;;
;; The jde-javadoc-pretty-clean function can strip formating information from a
;; Javadoc entry:
;;
;;     /**
;;        This is some example javadoc.
;;  
;;      */
;;
;; You can then use the jde-javadoc-pretty-decorate function to add formating
;; information back into the buffer:
;;
;; 
;;     /**
;;      * This is some example javadoc.
;;      * 
;;      */

;;; Code:
;;
(defun jde-javadoc-pretty-decorate-buffer()
  "Decorate the whole buffer."
  (interactive)

  (save-excursion

    (goto-char (point-min))
    
    (while (re-search-forward "/\\*\\*" nil t)
      
      (jde-javadoc-pretty-decorate nil)))

  ;;need to fontify when we are done.
  (font-lock-fontify-buffer))

(defun jde-javadoc-pretty-decorate(&optional fontify)
  "Make sure that javadoc has * characters before it.  This should only be
called from within a java buffer."
  (interactive
   (list t))

  (let(start end)
    (save-excursion
      (save-restriction
        
        ;;get the start and end and then narrow.

        ;;determine the end of the javadoc region
        
        (setq end (jde-javadoc-pretty-get-end))

        ;;try to go to the start of the javadoc.
        (setq start (jde-javadoc-pretty-get-start))
        
        (narrow-to-region start end)

        (goto-char (point-min)) ;;because it is narrowed
        
        (save-excursion
          (while (re-search-forward  "\\(^[ ]+\\)[^* ]" nil t)
            (replace-match " * " nil nil nil 1)))

        (save-excursion
          (while (re-search-forward "^$" (1- (point-max)) t)
            (replace-match " * "))))))

  (jde-javadoc-pretty-indent)

  (if fontify
      (font-lock-fontify-buffer)))

(defun jde-javadoc-pretty-indent(&optional end)
  "Indent based on the start and end of the javadoc."

  (if (null end)
      (progn 
        (setq end (jde-javadoc-pretty-get-end))

        ;;add one line
        (save-excursion
          (goto-char end)
        
          (forward-line 1)
          
          (goto-char (point-at-eol))
          (setq end (point)))))
  
  (jde-indent-java-region (jde-javadoc-pretty-get-start) end))

(defun jde-javadoc-pretty-get-end()
  "Get the end of the javadoc."
  
  (save-excursion
    (search-forward "*/" nil t)
    (forward-line -1)
    (goto-char (point-at-eol))
    (1+ (point))))

(defun jde-javadoc-pretty-get-start()
  "Get the start of the javadoc."

  (save-excursion
    (search-backward "/**" nil t)
    (match-beginning 0)))

(defun jde-javadoc-pretty-clean()
  "Strip javadoc beginning ' * ' should become '' "
  (interactive)

  (let(start end)
    (save-excursion
      (save-restriction
        
        ;;get the start and end and then narrow.

        ;;determine the end of the javadoc region
        
        (setq end (jde-javadoc-pretty-get-end))

        ;;try to go to the start of the javadoc.
        (setq start (jde-javadoc-pretty-get-start))
        
        (narrow-to-region start end)

        (goto-char (point-min)) ;;because it is narrowed
        
        (while (re-search-forward "\\(^[ ]+\\*\\)\\( [^@]\\)" nil t)
          (replace-match " " nil nil nil 1)))
      
      ;;use a custom end... basically this is the first instance of a " * "
      ;;after the start.

      (goto-char start)
      (re-search-forward " \\* " nil t)
      (setq end (match-beginning 0))
      
      (jde-javadoc-pretty-indent end))))

(provide 'jde-javadoc-pretty)

;;; jde-javadoc-pretty.el ends here
